from __future__ import division, absolute_import, print_function, unicode_literals

import nose as __nose

assert_eq = __nose.tools.assert_equal
assert_ne = __nose.tools.assert_not_equal
assert_lt = __nose.tools.assert_less
assert_le = __nose.tools.assert_less_equal
assert_gt = __nose.tools.assert_greater
assert_ge = __nose.tools.assert_greater_equal
assert_is = __nose.tools.assert_is
assert_is_not = __nose.tools.assert_is_not
assert_is_none = __nose.tools.assert_is_none
