# AUTOMATICALLY GENERATED FILE
# DO NOT EDIT

# Generated from network-wired-4.svg

icon_network = b'iVBORw0KGgoAAAANSUhEUgAAADAAAAAwCAYAAABXAvmHAAAA'\
               b'BHNCSVQICAgIfAhkiAAAAAlwSFlzAAAOxAAADsQBlSsOGwAA'\
               b'B9xJREFUaIHVmG9ME2kex78z02k7lFJLcdtyePYqrBstkJx/'\
               b'zngkZjG35s6IqXqrb0z0jQlB1BiJutlEvHvjG194ycVEX+Gf'\
               b'240bMFFXUZQIJhKWhSCtEIOUtiwtUO0f/pTpv5l9ITPbwgC3'\
               b'Cq3+kifzzMwzv+fzfX7Pb55nBvjEjVhm/2RfX98DhmH4mb5I'\
               b'kiQJoU4QBAGAEI4zheN5HgB4nue5maNYF87j8fjdoqKi/8qW'\
               b'k37btm1yuVz+VVZWFgiCAEEQIElSrOMdfcr5DPmCJRaLIRQK'\
               b'vQYAcjkFCDYwMIC6uroUeJIkQVFUSpHJZGKdJMmUQhAEXr9+'\
               b'jbq6OgBALBYLp00ASZJYu3btvPelopBcF4yiqDl+lnUKCWY2'\
               b'm7FmzRrwPC+Cza7PPgr15OtmsxkmkwmxWCy9ApJNCryhoQG9'\
               b'vb1im8rKSuTm5s4rJNnSIkDoeDb4rVu38OTJE/h8PgwPD4vt'\
               b'HQ4HCILAlStXUpJ3tr+0CBA6JwgCPM/DZrPBbreDIAg4HA7Y'\
               b'7fZ3ILLfUIRo3LhxQ4Q9cODAHCFpESCY0HFXVxeuXbv2G4BM'\
               b'lpLAyXlw9epVsd3+/fslo5G2CAhGEIQ42lLrQjIkx3Fz/GRc'\
               b'AEmS4qgLa4EgQmjPcZx4TCQSkgIEcWkXoNVqUxas5IUrWUAi'\
               b'kUA8HgdJkigqKpIc/YwICIVCoGlaXHllMpl4LrRNho/H43A4'\
               b'HB9HBHieT9lC0DQNuVwOmqZB0zQIgkAikUAsFgNJkuKzhYWF'\
               b'mc0BjuPERM3JyRGnkG4FB4WCh0JBQC5/t+dhIyRYlgPLRpHN'\
               b'cHgTkMHlcokjnvYICAkp2OTkpBiBPxpDkMujYJg4lMoEKIrE'\
               b'2yCF8FQUkWgUSnoKgXFNSjQEEYJlZAoJb6EV2SxksjgUigSU'\
               b'yjhcY39CaDyCyUkOkQgHi9kHispN8ZOxHBBW44qKCuzduxeA'\
               b'9I5zth2aET85OSn6S14fllTAnTt3/lNRUXFMOC8oKJAn5wAA'\
               b'sCyLSCQiPjOfCKm9j9RudckEvHjxIlBYWKhpbm4uLC8v/8fN'\
               b'mzfzGIa5lQyRvCcS4KV2mFJC5mu3JB803d3dfovFsiIQCBAm'\
               b'k+nvzc3NP1gsFofVav3S5/PB7Xajvr4+RYgQGSHJpeZ48j0A'\
               b'KX6WTEBHR4e/uLhY6/V6EYlEwPEcJj6f2Lfyi5VqANi+fTuG'\
               b'hoZgMpnm9SElYL52s/180BR69OjRNEVRSo/Hg2g0Ch48OmWd'\
               b'+IX7BWVkGaIJQE4BBw8exL1798REfF/T6XTQ6XSYmpoSr733'\
               b'b5WHDx9O6/V6ZSAQgFwuh96gx9Ppp3DzblR/Xg2vLw9NfUD1'\
               b'lwBNvXvm0qVLj27fvr37g1TMWEtLSxQA914RaGxsFOGDwSDG'\
               b'x8fR3tMO/xY/rNlWvHjJ4bmHx74/E6AjYYDj8PTnn8dfvXp1'\
               b'oKWlhV0KAYL9bgGNjY3TBoMhBb6rqwuRSARff/E1RsJ5eDJI'\
               b'Y5dlDGvVKqC7B2GNBi6X66fLly8HlhIeAKjfCR82GAyMFHxt'\
               b'bS0sFgveeAewKleGwqxxkD09SOh0GFWpYDKZzOvXr//L3bt3'\
               b'/5cRAYvB6/V6AMCqPxgx1N+FgtFRRDUaBPPyAAAKhQIbN24s'\
               b'DIfDgfb29valEvB/vUbv378/LQXPsizOnTsnwgOAx+PBd99/'\
               b'jx89HoTz80V4o9GIhoaGty0tLd8tFTywSA5cuHBBU1xcPJKf'\
               b'nz9nzrMsi9raWhgMhhT48+fPQ61Wg1Aq4fV6sXr1ahiNRly/'\
               b'fv3NoUOHVi4l/KICsrOzV9I0TSfDC6WgoABarVYSvqSkBCqV'\
               b'CqFQCC6XC48fP/YdPnz4s6WGBxbJgdLS0q9cLtd1pVL5T5Zl'\
               b'yampKbAsC41Gg8HBQfT09GDr1q0YGxubA88wDFQqFex2e/jI'\
               b'kSPahfr5EJt3IaupqVmbk5PzE0mSlNVqVdlsNgwNDSEejwN4'\
               b'96HS39+P/Px8+Hw+SfiXL1+Gq6qqVMsFDywQgefPn7/dsGHD'\
               b'CoZhtnk8HqKsrAyjo6MIh8Pivx29Xo/BwUFoNBrJkT969Oiy'\
               b'wgMLRODUqVM/kiS5avPmzUaPx5PHcRyys7Ph8XjAMIz4IU5R'\
               b'lAidbnhggdcox3FKrVZbHI/H80pLSxGLxTAyMgKXy4VgMCj+'\
               b'UcjKykqB7+npSRs8sMAUamtru7Zly5a9kUhE7/P54PV6oVAo'\
               b'oFar0dHRAZVKBYPBgKysrBT4Y8eOpQ0eWCACp0+f/ra8vHyd'\
               b'Wq3G9PQ0aJpGb28veJ7Hpk2b8ODBA97v9ycyCQ/ME4Gamppv'\
               b'd+zYUet0Oqm+vj4Eg0HodDrwPI/Ozk6/2WxWGI3GqdbW1gPF'\
               b'xcX7BgYG2OPHj6cdHpBI4pMnT36zc+fOfzmdTsputyMYDGJi'\
               b'YiLOMEz7unXr/mqz2V75/f4fSkpKKp89e7bf6/XGnE5naybg'\
               b'gVkCTpw48c2uXbvmwI+NjZ1pbW29uGfPnmcEQYzW19fvy83N'\
               b'/Zvf72/KFLhgooDq6uqzu3fv/rfL5UqBHx4ePtPW1nZxphkF'\
               b'IJEZVGmTAUBVVZUk/NDQ0Jn29vaLSe0/KngAkFVWVp61Wq0p'\
               b'8OPj4wm3232mo6Pj4uIuMmtEU1NT3O12i/ChUCjhdDpPd3V1'\
               b'ffTwACCbDe9wOE53d3d/EvAAIEuG7+/vP2uz2T4ZeAD4FSh8'\
               b'Mvy2tkwwAAAAAElFTkSuQmCC'
