# AUTOMATICALLY GENERATED FILE
# DO NOT EDIT

# Generated from media-playback-stop-4.svg

icon_stop = b'iVBORw0KGgoAAAANSUhEUgAAADAAAAAwCAYAAABXAvmHAAAA'\
            b'BHNCSVQICAgIfAhkiAAAAAlwSFlzAAAOxAAADsQBlSsOGwAA'\
            b'A49JREFUaIHtV7uOI1UQPaeq29aQzNg9ngzxAStCUkKQSFaj'\
            b'1e6QLivET/AJfAQZEhFC/AAfgSW0KULaHUjAmgnWt4qgbt/u'\
            b'9gxi20tP1Eeyum33vVXnnno1MGPGjBkzZsyYcTQ48vkTAA0A'\
            b'PdbgcrkkAC6XS1kul3R3LhYLqeuaImInJye/b7fb3dvuV420'\
            b'3zx/8fyr3e6vr0eueyusVs03L399+e12u/1liv0B4IOnV0/8'\
            b'9etX/n/j+o9rf3r1xC8vLz8a49BYBQAAm80FfvzpB6gqVAQq'\
            b'ClUFRWGWsN/vYZZAEgAhwnyPcgUJgjBLMDN8+slnAIDz8/N6'\
            b'cgIAICL5o6AIKIQooVqDQqQ94Q6IdI4PSRAkIKIw25d9z87O'\
            b'piNwenpakreqKpCEUCDCIEMpxPaUgQqHCnRXwGxRbDRNs8AI'\
            b'jCKwWq3K6agEF1HpqcFyJYmUQoU4/aHz7vEbAIh0NpqmmU6B'\
            b'9XpdTqeqYinBHoFOhcVCsN8LUko951slPK7OO4V8UgL9+Kwq'\
            b'hXvcDwnksFKCFKgKzPxuEiPCJxRC+W/SJL64uKiTvwEAqFYA'\
            b'vPzXEdASViqKlBT7tAdxEEJxl/foSK3X60lDqL7+89WAQLjg'\
            b'AwVIKcnNnBOW5SohBBbX/cDGZASapikEKhWAhHt3luG0BpFW'\
            b'BRCJCvME9y6UIv79XhuTERhUIa3gWYGWiGYVeJgTKlGRzGJx'\
            b'IeGDQwCAuq57NWlCAqIyjOA2vvvO56pEIYTRpYuzeXF8u6vE'\
            b'JASapqm7RBRIZCYIFoco95XVGCfMCDOL53IF8o7F9AT6JU4k'\
            b'yiVyjW9VIGI8EJVep457lyBg3jnt7oMQmpTAarXqtXmW0yZK'\
            b'IYcw1Og73nZnQIOApTh9BwCHuz0Mgc1m0ynA6MLtkMYcSmzD'\
            b'ikFQRQsJ5K5tJjAzuDvMDfSj34/G94HuWzsGtAQkRgZ0s5Bm'\
            b'BVRbEjFOmBlSSiUfHiwHBmgVQJ6HeKBGeRfoh1GoMGhfBtg7'\
            b'hNComnsI/9ej68bnQX4yVoVCMpiLjsXRCrg5wLaCeDS1nMDx'\
            b'e37ODdG/fNADyrqHrEIDAt4aN5gR0pLJnY0Sp59y84r5vyNA'\
            b'AJb3eJcQOorAbvc3Hj368Gij9+H29vaoZB5L4Obx48vvX3z5'\
            b'xbPxpv4bV88+/w7AzZg1o7LIzN4H8DGA98asG4EbAD+LyG8T'\
            b'7T9jxowZM2bMmNHHP7fujLaX/o8yAAAAAElFTkSuQmCC'
