# AUTOMATICALLY GENERATED FILE
# DO NOT EDIT

# Generated from document-save-all.svg

icon_save = b'iVBORw0KGgoAAAANSUhEUgAAADAAAAAwCAYAAABXAvmHAAAA'\
            b'BHNCSVQICAgIfAhkiAAAAAlwSFlzAAAOxAAADsQBlSsOGwAA'\
            b'C5tJREFUaIGdWc2PZUd1/52qe9/tnukesD14ZrqZwcYwgQBB'\
            b'iieCVcjSiySbSLZYINYY1pZAsvkPQIiNFcOCKAEJdpaibCJZ'\
            b'WUAiI8gQnFjEiZ3YHo9xf0y/fv36vnurzjks6uN+vI8xVKt1'\
            b'v6vOr87v/M6peoQN7Yknnnjus5/5k2+23pn+/fnsDH/zxS/B'\
            b'lhNURRHv6uhr6o4EEBEoHgdvEEAgUP968Cmhbh3q8zP85Id/'\
            b'j53d3fCICB+6fPmgwIZ2dHBg7t27Z/79V7dR2AIPPvggbly/'\
            b'jqZp8OEbH8HF3Q9iZ3trUxcjQ5EHD8cRiJ7xRBEUASdnNU7u'\
            b'HaNpGhS2M/nYHH9oIwAFwXuPajLBww8/jOl0ClWFqobODWAN'\
            b'beoiG7oSDI1nngbPKQML4FQB1c7TqsBmANHYBx54AFevXkNd'\
            b'1xARiAqIAEMUAYzpMzB3MPt9QAMQ1D2jRLt0HZ+pCESks09k'\
            b'MwBAAQLevnMHZ7NZD1QY3hDA4tE6WRsB/SvqXdLoRerdyyAA'\
            b'VJXNXkoTmppsAGAB3Do4PLx+eHiI/b393PG7v30XTdPi9i9/'\
            b'gYu7O7i0vbV5/lcxjJbBEAjGWvzZrVuwhc3PFcAMLQQa/gYU'\
            b'UhTPPvvs4dnZ2UP9/sUzHnns4/jkZx/H9qSEoX53A/PyOVFS'\
            b'm2WlydzGkCrJeFFg0Tj87Kf/gh//6EcwhQUAXNjexl/+1V/j'\
            b'kZufAkShGMeAojh87+ChN954A9PZKdh7PP6nj8N7j93dXTx2'\
            b'8xPY3d5CYe2aOe5BoeF5UpUOBAZ8z8AIcJ5xMl/g1Vdu4/z8'\
            b'HEWU5sV5jYPDI9z4uEJUgaUgVhSqiqZtsLe3h3fu3Ol4FugP'\
            b'Y+j3V5pMERoF5whEAmo6D415boyBiEIEwKoYEBGU5QR7V6+h'\
            b'nkeVEYEiKI01BDtIYyvnf6jz/esRffK9nsoYDhTUaFRfaZgl'\
            b'AFBdeiYiwQN1XeOV/3wF05Mp9vf3ENgWOrUE1I0Ls7PadHTz'\
            b'PfRC//FY38NlONmqinBPlz3gmSEaAlhEVlPoj27ezA+STBI0'\
            b'zI4BPnCxep9KQ/3TFYHbAe0DXLScDR8DUAVYNHphDYB+C/xX'\
            b'IAMgmE0xMEpSGcRSUPfGAMKsxhlnlXi9DEC0Dw7LMTAGkEEg'\
            b'ZFpjCPNF2408Mn7FaeBz9mj3ocbrdIwPUZYWIlgJwHnNHtB1'\
            b'FFoFQNB1uHthssLM0Xfa3dFoueb7mp/3gYXEBLSO4/1lAASJ'\
            b'5cuaGOhHdWoiAuEQ+cwKb5bfWWd86rgDcn8AXqPKrFAaFgke'\
            b'0OiBcS20zgOqDFHAs+B84VaXBDo49IzWwY3sGR15J75QFjZM'\
            b'lqyiUJp95Fjpg9sAAFBReBFUk5AZ+8qx4rPOqFUeWAEgxYfz'\
            b'ChWF9uKg36tEYOnYt3N9DGhHIcJ9KLTO+J6RmwAEigDaU5zU'\
            b'/B8axBrdxSKoF+79GT8+0zGQIcUSgCJSSGW54hSNmXhFgG/2'\
            b'gIQBHAuKMhRzGzRoYOXY6EyitR6Ik4Vlrb/983/Df73ya0zv'\
            b'Ha1JZCtUKER7UAQWWcn3+xm/BGDkgSEAhFlmBUSgvWCbT+/h'\
            b'7f9/HaqKnQsXByokzChkhXWiClYOadwLaudXq9DI4nXGL50N'\
            b'AABFYQI9olxSz6ayLHFp9xKcc3n2k8zeJwZCZ14E1m4up8c9'\
            b'jLtcFQ/Uu8/RoESRYRAzFosFxnam99bHAMd6hVeZeB8wA690'\
            b'3OkA9IM43DeEQT2UGgujaRqICCaTCYwxg5JjQykh2QvvKwZW'\
            b'GL8UvNHqJYkNWxxBhUYADg4O8Nprr0GhMMbg03/8qfxsbTEH'\
            b'AP/8T/+Iw6NjnJ1OsarceN9tRbYeN2MtdnZ2cfvlf8Xu7s4A'\
            b'wOXLD6GqKjAz3rn7DsiYbLz0Vai/ELfGYOfCBfzmP37Z2ZHd'\
            b'Pjymb4nWx0neDNvwnqri4oVtGKJhvaMKaw0IQFlOwiZArImE'\
            b'eegBa21eUJdlOQCW7nvv8wwQ0aBC7Bvap0I6TiaTAQATZzO9'\
            b'61yXMHO/AnjnwSJwrkXbtrk/W9gAwPTckjrp9i8pG56MzusF'\
            b'Ywb/aVAiytf999MYyWhmzt+syrIAoKSwRQEwY3trO9vDzBCO'\
            b'McDM+YG1FsaYbAQA1ItF3tbrG5iMGx/T4GVZoixLqOrAc0VR'\
            b'DACISLZhe3sbzNwBYAV7D9/77+eDTKHEzz4NyrLEwdERFIrH'\
            b'Hv1o3jUIMsvw3kcNVxgK8maMAQvDUPAKZaCCuFeAO2/fwZ24'\
            b'heOcy0axCD68v4fLD13OnnDegTl4n72HZw+K5YY1JmRia21v'\
            b'oyb8iyrqxQKubfH0176KK1euDCjQ90D/P+l0f1L6TVVx/fp1'\
            b'PP2VpzGJ3knvn85mEBEYa7vtHSKwKjx77Fzaxc7OLmZnM0DD'\
            b'jkWhImBVlGUZ18DdJlCSqqZpcHZ2BmMMvvOtb8N7n2nknAOr'\
            b'4Ovf+Eb+zvZ28pg5n6dYeeutt2AMobA20IUQ+FxYcKxaJNJI'\
            b'mSGisLaAKODaFsod5fKSUkSy8SkOgCCpaWYpeoasARRxwxVo'\
            b'myYbSkRgZlhrl4KSmVFVVRQKyisk9pwFxLUOTdPkb8PGVjC2'\
            b'XSxQWgtjLZxz3c5cGjgd+zPMIqjrGovFInQonIsuVUGzaHA6'\
            b'm2EWt98T+CS7Y4otFguUZYm2bWDTJEWVmUwqmEi7FGvsPdwi'\
            b'AErfGGMgzBBjVgPoH10btLeu6+AV1bBPJIAiJJ3UaV8Wk+yG'\
            b'tYXkHKMp5hDK9jRhzAxmDpScdP2EvdMw7qScoChLLOo6e7QQ'\
            b'5hx4ZA1UJXI17AmVRZENC6OGossQQcmgqqq80O4nptSICEVR'\
            b'ZGoBCEZ6Bhsf6y0BQBDv4FsHXzmopBqkmwz2PoNPQV5wWkAQ'\
            b'oD7lg9CpEQXFQE1SJ3GP0osARDg7n+NsNsN8Ps/JK1EpDeZ7'\
            b'A6eZM4agJv5IIEDbtiBrsXPpEsqiQONaiGeYwqKclFBRNKpw'\
            b'MZ+wCMAcKJQzpjEwhQ1uU4ARCnfnXBiACNYWKMte0iMD5pCc'\
            b'+kVfSjrJu6kE0Kh4ihhjLJFiBUQYLAwrBuzChLm2zZRpnUNh'\
            b'bKal6auQjdFNoJCUEJIZNABomgYAMJ/NshySMajrc5yfzzGL'\
            b'Gp7kLdGlKIr8fhoYANpmEX55zFslSdFalPGn1KBMjPPzWtO3'\
            b'HKRXAcXJ6bTNMQBV+Lbt4oEoBKftEpOqotqqIBwqQde2mE6n'\
            b'mJ5M85IvUcfGZJQUjUbqQhqNF4GPFJWoOq218LG/h69c+bvJ'\
            b'9tZ3i6Kg9957j+fzeTudTv3R0ZEHcFowM6uqTQFHROBY0Im1'\
            b'4F7toaqYTk+hABaLGm3bBn4XQZdTDZPUZlzUpT5EBKwCG8sR'\
            b'ovADH4mgiLV/8uDjt269+7nPf+7V6XTqnnzySYfR0qL4nzde'\
            b'//KjNz7yA1W1/SKuq2HCgCkP/PlffAFQBVGoefoFXmrOuUyj'\
            b'fh+pkAOAtnUwZLqSQRXeOUyqqsvCqvj+97/XPPfN5+ZY04qj'\
            b'4+N/UGDrkes3/rYoCuMj8slkkhWDvUdd1yiKAnt7ewCCZKbZ'\
            b'A5BjJMVBolDKyMaYvMYoigJlWca4i6U8h3ebpgnfRKXJcroO'\
            b'AACcnp6++s67d+ePPfLobs6iRLEjhbEWVVUt1UkpOPst6X5/'\
            b'cdIvzYFQH7m2iYHcVcHM3UoreZWZP31fAC+++OIvnnrqqW/9'\
            b'9+v/+/X9q9fiD38ljDE4PT3ln7/8sr167RqZyNV+tQkAKkIc'\
            b't/ISHay1qj1AuUWvnjeNVEVpnGtzjHAUFCKCc06JiE5ns2YT'\
            b'AAKAl156qQCw9cwzz3zMGGOJyFZVZVSVnn/++Qf39/dvqurF'\
            b'qqoq731JRJWqbjFzyczWGGMBWGYmAH4ymfi6rnlra8stFotG'\
            b'RPz29rZYa0lEvHNu/uabb/7mhRdeOGmahk9OTtqDg4OaiOzd'\
            b'u3fb2WzWHh8f+7ZtFcD/AfDrAPwO3Nl7rmHsMvwAAAAASUVO'\
            b'RK5CYII='
