# AUTOMATICALLY GENERATED FILE
# DO NOT EDIT

# Generated from pictograms-protections.svg

icon_glasses = b'iVBORw0KGgoAAAANSUhEUgAAADAAAAAwCAYAAABXAvmHAAAA'\
               b'BHNCSVQICAgIfAhkiAAAAAlwSFlzAAAOxAAADsQBlSsOGwAA'\
               b'DPZJREFUaIG9mnl41OW1xz+/32yZmUxIJpmsZCGBbIQEEmIg'\
               b'i4oUWVSqpU9xQVEQ3IrtFb1qvdU+vVKlovQWZCmVqmAFBdpH'\
               b'qyxCCWYBQgkhYQkhIZlJyD7JJDOZZJbM7/7BIiJJZqD3fv+b'\
               b'mXPOe77vOe97zvu+I/DvQQQwQxCETCAdiFUqlSE6nU7ndrvd'\
               b'DofDMTAw0AoYgTOSJB0GvgUu3urAwi3o6oGFgiAsSEhImDR7'\
               b'9mwhZ0o+kyZlEhMdiUajpn/AgUwmolDIsdnsmBovcrLiBEXf'\
               b'HmT//v1SfX19mSRJ24EPge7/LwIRgiD8V0hIyOLHHluoWvLU'\
               b's+j1esqq2qiut1LX1ENLex+9fU48HunSIAIE6vyIMGgYGxNI'\
               b'cpw/2WnhNF1s4sPNm/jLXzbbrVbrB5IkrQDa/q8IKICXQ0JC'\
               b'Xnvl1df8Fi1eTGl5C7tLLnKmzowk+TIsiKLAhMQQZuVGkD0h'\
               b'nI0b1vP2WytsVqt1BbAKcP87CaSKovjpwoWPp6/8/SoKj7Ww'\
               b'bc8FunoGfPN6CAQHqnnknnhy0kJY/sLzbN++/ZgkSQuAmpF0'\
               b'vSHwE71e//Enn36mDR+dxOotpzG1WG/d6xtgzOgAXlo4gYry'\
               b'Ep5eurivt7d3PvDVcDojEXg6JSVl3T++2iscOtHNli/PI/mQ'\
               b'K2o/OS6XB/egx2sduUzkifuTyBirYM6sGYMmk+k5YONQ8sMR'\
               b'eC4jI2Pt13u+YcPn5yk63jzi4IE6FVnjQ4kM9Sc/M5LEuCBs'\
               b'dhf7So0cP93OyeoO+vpdXhGZmR/DgtmxTL/rdqmmpmYp8Gdf'\
               b'CPx43Lhxuw4VlYqrt57lWNXIG4MgCLz5/FRyJ0UOKePxSNSa'\
               b'LHz69TkKy5pGtFmQFcWTD8RxR0Gup6Gh4QHgC28IpAQHBx8/'\
               b'UnZcvXV3M4eOjVxrBAGeeGA8j85N4bzRwrjYwGHl7f0ufvbC'\
               b'115FY2ZeDPfmBTI1J8tqs9kmc93Cll0nrxRFcfe2z3ZGn21S'\
               b'8sXBhiENq/3k3DE5iuk5MTw5L43pU2M4dd7M878rZGKSgdBg'\
               b'DcIQ8RVEAZVCxrFTI0e2rrGHMaODuefuPNWuXTtygQ+Aqwvx'\
               b'egKvLlq0+OG75/yMVR9VDru3L543nmcfyiA9KQSDXs2Fxh5e'\
               b'ea8Y+4CbPcUNtHT0kTQmCK1agSTBjr01pMTrGfRIfPrVOR6c'\
               b'k8zpOjPN7X0jkjhRbebJh/JoNNZGVldX24DSq5NxjVxkUFDQ'\
               b'+epztZrlq8poarMNaTA4UM3WlTMpP9POnmIjTtcgp86bf5AS'\
               b'oiiQnRZG0hg9FWfbEUSBu3Nj2Fti5LWnbkOrVrBsxUHqm3pH'\
               b'JJEYF8QbT6WTlBjfZ7PZxgAdcE0EBEF4680Vb+WZHaH8s2z4'\
               b'HWf545kEBfjx0qoiak0WLrbZcLl/uFVKElxss3HyXAdtZjtt'\
               b'nXZKT7TQZrZzuKKFgswoZhfEUXisCXv/8IXXbBkgPnoUkQa1'\
               b'sri4WAbsg+8ioA8KCmqqu2BUL3q9EEuvY0hDYcEa/vrObF5+'\
               b't5h/nfapbfkBRulUrH75DtyDHpatOIjDOTisfHiIlrW/mkpc'\
               b'bLTdZrNFAL3i5d8ef3LJEnVJefOwzgMoFDIsVsctOw/QY3Xw'\
               b'5sajjI0JJCc9fET51s4+qmrMzJ8/XwMsABABBEF4cOHCxewp'\
               b'HblYdXTZCdSpiDBob9H9S2i83JaMNHFXsLukhaVP/RxBEK4S'\
               b'iE5MTMwOMYRypq5rRAMO5yB1jT1kpYbevNfXICxEiyCAqdW7'\
               b'/qr8TDspKUkYDIYcIFgECqZPn86/Trd73eecrjUzISnk5r2+'\
               b'BlGhWiy9Dq8j4B70cPxUMzNnzhSBH4mCIEzNv30a1fXed5gl'\
               b'5c3ERQbcpMvfR2iwhvqLPT7pnDPayZmSiyAImSKQkp6eQW2j'\
               b'90YuttvQqhWkJuj5yYyxTMuJRhiq7A4BjVqBTCag0yhpah26'\
               b'5twItY0WMrOyATLkQFzM6EhaO6q9UlYpZfz84QwCA1Q8fE8y'\
               b'sVE6woO1vPhEFvZ+F34qOe1mO0cqW2jttOPxSKhVcnT+StRK'\
               b'GYOShEwUyZsUgVatYHdRAxq1wicCze19xMYkA8TI1Wp1uEar'\
               b'psfmXQ6OHxtMSoKel98tJic9nNsmhLNpxynUfnKsfU4czkHU'\
               b'KjkP/CiB3UVGosK05GdGsXPfeZxuD6F6DSFBajR+cqrru8nL'\
               b'jEQhF9FplVj7nF75YLE6CNYHIgiCQa5UKpVOp9vrM22v7bvD'\
               b'+qNzU5AkCad7kGfuT78q887m4wQGqPDXyAkO9ONCYw8qpYxZ'\
               b'+XF4JImzdV2s21ZF/OgAvjh4gfTEEB65N5kN2yu98qF/wIVK'\
               b'pUChUASIMplM7sspq67RgscDXT0D2Afcl3P/+/k/SqfE3u9m'\
               b'w/YqPttzHlu/i3CDlsAAFfpRfuROikApF3G6PaQnhmAfcNPZ'\
               b'3e+1D1fWm9vtdolWq7XXT6VEJvNuEUoSFJVfJDVBT/flQ31L'\
               b'R9/VqAC43B767JfSwd7vwl+joMvy3QVAm9lOZJiWPruL88Zu'\
               b'bp8cRcmJkYvoFSgVMtzuQTwej1N0uVydtj47Oq3SawNV5zqZ'\
               b'mGzgV38ooayqlUnJBs7UdTHgcNPUZsPpGry6q5larCTGBlFr'\
               b'slzVrzX1MDHJQGZqKHWNPeg0Slo6Rm6rryAkSE1LWwdAiwiY'\
               b'jKYmokL9vTZQWdPJ5LQwTC1W3vuwnJl5sTicg+wtMVJrtPDT'\
               b'GWNZv+1SPtvsLo5WtjDn9jGs+aSC1R+VU366HbWfglfeKyYz'\
               b'JZTq+pE7gGsRHqzBZGoCMMqByooT5dPio5M4XWv2ykBXzwAO'\
               b'5yAJ0aOoa+zhwRe/Jis1lFE6FR5J4u0/H/teZ/mb948wfUo0'\
               b'd+fFEqrX0Ga2s/KDYwCkjQv2mcDY2FGcKC8BOCOXJOn44dIi'\
               b'5j6YxZcHvTdiarHym+emUFbVxppPKiitaBlW/sCRRg4cabz6'\
               b'OVCnIm1cMFnjw/iysN4nAsmxWtbvOoQkSaUi8M2ePXuk7AkR'\
               b'PlXT3647yqYdp5hze5xPg1/BOy8VsHJ5AWfqzJy94H0EFHKR'\
               b'ianhFBYWeoASEWhtaGgob21tZUJisNeG3IMeio434x6UMOjV'\
               b'PhMwBKnZvOs0r64uYcDh1TUoAJPTwqioqKS9vb0EaBMBJEna'\
               b'8qcNa5mTH+WTE5IkUdPQzcRkg096CrlIgL+KouO+Pw/Myo3g'\
               b'w81/4vK1PFdOZB9v2bLFcduEcJ9n8+yFLrLTRj5NXYvIUH+s'\
               b'fU46uuw+6Y0O8ycxNoBt27b1AlvgOwLdFotl0/p17/Pw7ASf'\
               b'jB471cbEZMOQd0A3QnJ8EPVNPT5fyT96bwLvvfsODodjE9AL'\
               b'3xFAkqQVq1b9vn9qhoGE6FFeG60810lXzwA/nZnolbxOq2Tu'\
               b'tARO1nT65Hxqgp5xMWrWrPmjRZKkt698L14j09rd3f3f//HL'\
               b'ZSxfmOZDayHxxtrDzMqPoSBr+DUkCPD6szmEBav5+/46r51X'\
               b'KWUsX5jGc88sZWBg4E3gKvvrvZSLolj2+Y6/T3Iq49n4+Vmv'\
               b'BwnwV7Lu13dx4IiJLV9W477unigwQMVrS28jMEDF62sO+9Q6'\
               b'/OKRNFovFPPsM0sPS5JUAFytkjea5uSAgICysmMndDsLO/im'\
               b'tPEGIjeGQa/mPxdNZnS4P5/vPU9VTSd+KjkZSQbmzRjL7qIG'\
               b'Nv/t9A/IDYf7p48hb7yCOwqm9trt9kzge6Eb8no9Pj5+V+G3'\
               b'JeKmXRe8ehu4FuNiA5mSEUFspA6NnwKr3ck/jzRytLLVJzsz'\
               b'cqN5aGYUuVOyB1tbW+8B9l4vM1yiL0lKStq4/8Ah4ZPdRvYU'\
               b'm3wa/FYgCPDju8ZwX34od00rkIxGo88PHFfwdGxs7Nqvdu+T'\
               b'VV1w8cHfzvkU/puBUiFj2UOpRAQ5mXvvrMHm5ubFwEdDyYtD'\
               b'/XAZG4xG4/25U7L73JYq1r6ay9iY4R8vbgUTEkPY8Ho+7caj'\
               b'FOTlWJqbm+9jGOfB+2fWREEQts6bNy979R/WUlFjYes/6mgz'\
               b'+1ZJh8LocB2P3ZdAUoyWp5YuYu/evUclSXqE6xbsjeDLZY4c'\
               b'eEGr1f76heUv+i9b9gsqa8zsO9xK+dl2n1NLIRfJnhDO7LwI'\
               b'kuICWfn271i37n1rf3//G8D/AF4ZvJm/GoQKgvCqWq1evGDB'\
               b'At3CJ5aQmpLMiTOtVBv7qDX10Ga209ndj8vtQZIkNH4KAgNU'\
               b'jA7zJz56FClxWialRnCy8hQb169h586dff39/eskSVoJeHeq'\
               b'ugUCVzAKeEwQhPmRkZG5d955p5CffwcZk7KIiR5NWFgwctml'\
               b'95MBhxOz2UJ9QwPlx8s4criUffv2ebq6uo5IkvQx8Ffgpl7P'\
               b'b4XAtQgH8gVBmAqkAHFAuCiKKrlcLnc6nT1cehIyASclSSoH'\
               b'DuDjbN8I/wuzJT0nIDcjfQAAAABJRU5ErkJggg=='
