# AUTOMATICALLY GENERATED FILE
# DO NOT EDIT

# Generated from appointment-soon.svg

icon_timer = b'iVBORw0KGgoAAAANSUhEUgAAADAAAAAwCAYAAABXAvmHAAAA'\
             b'BHNCSVQICAgIfAhkiAAAAAlwSFlzAAAOxAAADsQBlSsOGwAA'\
             b'EFZJREFUaIHdmXmQXEd9xz/9jrln7x2t9tC50q5kGR1GRgLL'\
             b'NsKyiQHHZSSgMOJyiAMh4CQmRRIbCFdxmsJOChISymBkDDFQ'\
             b'IQRjrNjGlg8c2ZIsydJqJa2klfY+Z+ee1/3LH/Nmd/awjQMk'\
             b'VXlVv+qenvf6fT+/36/7db8Hv8dDPon1++wfQP0uOjmyi1gh'\
             b'xBss1BXKcV4nYpYabWoQCaKUVkrllaUmsOimaJ42yN5clide'\
             b'ez/Z/zMA+STWc6e42nacW7WRy2qWrMjH29fFw4lWFahKYIWj'\
             b'WE4QEUG8PDqbojg5RHrgnJk8eTiV6e91LdQD2nh3bryHxxTI'\
             b'/wqAgHr2Rq53XPtroYamuqatV8VjKzahTAFJX4DMCFKYBC8D'\
             b'xitdZDngRFHBaogkUNFmjMDkiWdl8OmHU97k8Dmt9S2b7mHv'\
             b'7xXgmXeyIhCw94Qbmi5qvfpt8VBiOTJ+DBk/AflkCa/sx7n+'\
             b'VJWlglAtqnY1qqaDTN9xen95f6owMfp4sejdtHkP/b9zgAPv'\
             b'5m3Kdr/Vds3OWE3nVkuGDyCjx0C0L1goZAqkx1OkkzkKOYOn'\
             b'DbalsBxLQiFLhWMhovVx3KADSpUo7SCq/iKoX8fYoYf1hUd/'\
             b'lqTgvWPjHn75OwEQUAffa38+WNPw4ZU7PxSzVRbpexK8LGDh'\
             b'FTQj54Zl9MI42kMVFq0lX9+BxBohEENMASufwZrqJzB6Anfo'\
             b'OG40JA3NNdQvaVCWbQEG3Ciq9XKKuSLd//qNlJ5KfmLjd83X'\
             b'fisAAfX8++xvhJva3rVi50eijB4qeV0pjIaBk/0ycm6Y3NLL'\
             b'VXbNW9B1K7CUheM62LaNskrdixa00XjFIoLBHjhK+IWfEu5/'\
             b'ThqXJUi0L1YKAwgqsR5TtZpT992Ryo8OfX3D3fq2/zHAwffZ'\
             b'n4k1td2ybNctMfoeQ6bOg+WSHk3Sc7BHMm2XqfSm92JHawlH'\
             b'IoSsIvbJX2FfeA7Gz2EmB1A1rTi770YAYwyFQoFcLkc2k0En'\
             b'+4n/+p+IjB6VlZesUMF4GKSIqlkJiUvpvvfLqfzI4N9u+I65'\
             b'8xUDHHg3bwvWJf5l1e6Px1T/PiR1HqwQg6cGZODMMBNX3K5U'\
             b'yzqqqqoInt2H2n8P+sIxGla24IYsArEwbjhI994DRD91FEFK'\
             b'U6pvxmgymQzJZBJ19mlq931VWte0UtvaqNA5VM1KTMNGjn/7'\
             b'86nC1MQfvvp7PPwbAxx8F8tVIHCw86bbqux0NzLaBU6E/q5e'\
             b'GRjSTLzxCyrW0EI8EoR730+McarihnhTHcoNQiCAcgOICEd+'\
             b'8gRVnzkOUBIugojBGIOIQWtNMpkk1d9D3S8+Js3L60isWKzw'\
             b'MqjEeopWA8fv/vKY0t6aDd9jaK7WeY96ASWuvafl6l1xW9LI'\
             b'6HFwogyfHpSBQY/km+5Q9S3LqUqdofjFLTRZ51icEGLVIcTz'\
             b'EO2B1mD0dJ+2XRoT02aVSsuysW2H6uoa6pZ2Mn7dXaqvZ4zx'\
             b'8yOCE0GGDuGGXVqvfHOVcp1/XMjZ8wAO7Oa6SH3zutrOS5X0'\
             b'PwFOiMx4ir7TAySv/ZKqb24jdHYfzv0fpLM9SCxqgechni9c'\
             b'e6A9xJsBcGx7BsKqgPDNsizC4RD1rSuYvPYr6vwLZ8inC2AF'\
             b'kAuPU7txu+PEaq7afyPbXhJAQFm2fUfrNW+Py9CzoIsIUXr2'\
             b'd8nklbepupbluH3Pwc9uo6065Qv1kErhWvswPlA5ApaNbTnT'\
             b'nrds3yzLN5uAG6BmaSfjW/5c9TxzXHBiUEzD6BHart4ZcwP2'\
             b'V18S4OCN7IgkWhtDDa3IWBe4VQyc6JFU6xYVWrmZ/b96CP3D'\
             b'j7KsvrQGE2+OcO2BHwnRfjrBguljTwv3TSksSxFwXaIbrmEq'\
             b'voyR0+cFJ46MHiW65CKceP2aZ3fz6hcFsALORxJbdsRl7Cgo'\
             b'Cy0uI6d7kW0fIhaN4oVq2H3oVTxwocEPmfHF+x73vPm/p8fA'\
             b'zDgoe9xSZfEWyrJQfhkOhSlu+6ga6DqjxAqD8ZDxEzReuj1q'\
             b'2/aHFgQ4souYoLZHl78KGe8GN8Zoz3nJLr1MRRuaUUpxxeXb'\
             b'+Itb/pSvdK/kY4fWlhi0Lgn1KqKgPfD0dAQce2bQWpZVyntf'\
             b'vKqAsJTCUgqlFLG2TrL1HST7+sGNIuMnqOnYrAzs+uEu7HkA'\
             b'BZdt0SWrPMtLQTGNcuOMnu1TbNxZ8pDf8datm/nAB/+Yw+kG'\
             b'tj9yKYfHorOF+2NAFhwDJc/PEm2p6b6Vfx+UwrJtzPobGDo7'\
             b'KMqthtwYtqMI1TdLe4BN8wCUo66qbl8XlXQf2C7FnEcxl5dg'\
             b'28UwfQNQKK68fBu7372beHUNf3V0HV860jaT9xWzkfgAZc/P'\
             b'pE2FaGbq5ZVqqQ2CHdvIDQ8pLTYoC0ldoK7j4pDYbJ8HYDnO'\
             b'peFEm0VmCOwomdERvMXrlGXZ/tNOlUwBSnHNNW/ghhtuIFLd'\
             b'wK/z7Vz38Dr6knbFoK4YA1Y5XdQ8K61Ky3fwneSDWG4Qr26Z'\
             b'ZMfHwIlCdphwotV1bOfSeQAipj0Qq0Pyk+CEyCbTmMZO5h4l'\
             b'/5SO6667lmvf9AcoJ0i0oYXdj3Xw3a5aX7w/Q8F0alR6eaa/'\
             b'cmX+fwC6YZXKJ1NgB5HcBG51PcDa+QBaqq1gCLwMygqQy+Sx'\
             b'61rnAcicncquG65nx44dFLTQuqSVH/W28Z6H2sjkNQNJmb5G'\
             b'pLSUKPfC3JrIwpvK2qXk0jlRdgC8NHYgghGpK//tTF9vTFDZ'\
             b'DmKKYLmIZ1DBWIXgipvKjCBBuPGdb8fzPB599FGqGxqZSkW5'\
             b'+t/C1IbA1fV85fQpli5ZUrGYY7pe2ffMgm+mzQrH0CMGlAvG'\
             b'Q9kWGInMA0ApQbxy3Y+kzBcsJaTKlaWI8L737KZQLPDUk08x'\
             b'ODZB2+Im6uuqGRwa5r1/dDPvufGdrFrVTnPLYhYlGuddPxum'\
             b'ImICWP7AECk9eyq86czoV3lTyEaUVWpSbhDJp6e9PHs5bBCx'\
             b'Zn4bg0HxJx+4ia6ubpJTKS7ZuJ7GxgQPP/KfxCIF9tx3H6FQ'\
             b'iFyuQCaboaG+gZaWxSxbtoTWlmYWLUqweHGCSCTsCy+BkJ/C'\
             b'Dbglh1oOUsihLJWZD2CpCZOdith2EMQjHAsjk30LizaCKIOI'\
             b'wpiKYa1g86ZLON97jmPHj3Hy1EkAbNvCdWxi4SAh1yEWCZAv'\
             b'5Dh9+jRdXScoao1X9OjsWMXtt906KyrWRC+hqphCF8AO4mWT'\
             b'KNTo/BSy6C5MDDWH3Sji5YjU1uL0HEVPe7lCtCiUKJQpTYGV'\
             b'88Zbb7iO7+65l3g0QzAYoFj0GBtL4rguoYjQ0JhgUaKRpqZF'\
             b'NNTVUVdfS21tDbW11di2jTZ6es8gItjDJwguvQjRWVQwTmFi'\
             b'CCyOzAcomqfTQ/3bIssbLUn2El7UjnpiH2I0Ys0WrVAYpWZJ'\
             b'9zMXy1L8w11f5W9u/zQFXWBqKs1nPn07Gzesx4hBjGDEYIzG'\
             b'GFMSrEulNhrxNzpGBOPlcUfPEG7cAWMHILaYzGB3wSt6v54H'\
             b'YJC9yTNdH2zs7Kii2IUTiRKIxqXYf0SZlvUoJRgjKExpPjdz'\
             b'N3N+pJSipXkx3/n2Nzl2vIu1nZ0Igqe9iu2kmQVhpFTKdHsJ'\
             b'Qp16kkhzmyglSnQRQnWMn+zOW3pmezkNkMvyhOrrdQ0uyglA'\
             b'dpjGzk6yB3+Mbr7YTxf/KTlHfFmYUlKKjG+rV7Xj+Q+zyrFk'\
             b'ygPfF69NBUw5AsbgPP8jEqs7FNkhcKN4+QLe5LBsWMXzM5nv'\
             b'H6+9n6xl8cBkz3FRkUVIuo/a1euVc+ZJkeTQdOfGv5n2w6+1'\
             b'rig9v+4tbMZDa40pX2M0WhuM1nP6N8jIKdzxHuLL1iCZflR0'\
             b'ERMnj2il1B71d5h5AADa8+4cOrg/TbQZvCyW5Fj0qo24j/99'\
             b'hYf80hdi5kLokkhPl39rPL9t+jwfRBs9ayzoir7dR+6Q5k2X'\
             b'CoVREI2EGhk68Fy2UNTfqNQ8C2DjPTzmTQ6fzQwPQKQRmegi'\
             b'seG1KjB4SDj7X75YUyGkUnylyJLHPd8qwaZLMzsSldGV4w8R'\
             b'KY5Su2aTkoluiDYxebZHpJh6dvO9HH5RAAWiPX3L+X2/ShFt'\
             b'BZ2H9AVW7niLCj34aZGpId9jC3l+JnU8PQdIe3hzIbU3+9oy'\
             b'yMgZwo99XVbsuE4xcao0fkIJ+p96LO0V9a1zZo75byU23cPe'\
             b'wuTY4yMnjmkVb0MmuglW17LkNVsJ3v9nojPj0zc2c9JmJnW8'\
             b'WUDTEAullN+P0RqZGiT8k1tkxet3KMdVyNQZVHwJI4cPFEw+'\
             b'8+NL7mH/ywIAFIveTQPPPJnM5xWEa5GhZ6ju2KRaL+og/IOb'\
             b'xUz0L+C9crp40yK9BdLIe5EomJEewvfdLEu2bCHa1o6MHIBI'\
             b'gmwyy+ChA+OqoG9ZSKu9UOO3DpN6f6c5NHWh9/r6tZcElE7D'\
             b'1Dkiqy5X4ZBN9hd3Ka9mKaa6ddZSeWa9ZBZYrBmMqXgG+FOl'\
             b'0Rqr6yEiD35KVm7foaqWr1XSvw8CETy7ju7/+Hkqncu/efP3'\
             b'6f6NAQD3n4/Q+472YjE/cGFL3ZqNAVWcguQpQks2qboVHaQf'\
             b'uVtU73OqmOhAArHZ7z2nIWZseu6XmamS0R4iD36K+IWnWH39'\
             b'jSpUXYP07wM3jA4kOPnzn6dGJzK3XvEDeZDScs7MFboQgEvp'\
             b'AefefUQO7lqRj+X7zl9cs/rigKU8GOvCjtXRuPlNKpAdkMLe'\
             b'byq77zCeE8XEEgjKF+wvB6S0dChHQLwc1uknCT1+F+ED36dt'\
             b'w3pp3f5WZaXPltImVIOnqjj5wC/S/WOpL1/1Q/OdCm3zIOau'\
             b'BywfYBoCcP/9rermpY3Rv1zx+iujwTCQHQEngmrchATqSJ3c'\
             b'z8Dhg+QGzyONKynWt6NrWiAQBRFUPo09eR57pBtr9Azh1mU0'\
             b'rd1AbOUmyPYjwwfAFCDcSHYqz6lHHk+fGsp/YudPzX1A0Tev'\
             b'om5eEQDg3vl6rrhsiXNn04Z10cZVKx0KE1BIQSCOqlmFqm7H'\
             b'FPPkhnrIjw6SS06ii0UAnGCAULyWYP0iQollKMtCJk8ik91Q'\
             b'zEAgjrhVDB/vKgwcOT71y9P65r9+nP0Vgn9jgFkpNLe+cyVN'\
             b'H36N89m6eGhL6+YNkfjiRshPQTEFxkCwGhVeBMFqcKNgBQAB'\
             b'Uyydk59EMv0lcMsuRciJMt47KAPPPZ8dTeUe+twT5rN7exl9'\
             b'EeHl+oum0EIQlaUDOF+6jK3bltkfj0bDyxs620O1S5otx7VA'\
             b'50AXS59XRfu7Kvwtqg22W/rkaoco5jXjZ8/rkWOn8ul8/uhD'\
             b'p/UXP/kkB32RZSsuUM46XuoTk10pusLK7fbnXsfGrW3WjTUB'\
             b'9UYnGjM1LYlgpK7aDcSjOIEglmOX9glFj2IhT2EqTWZ0vDBx'\
             b'YbjgZdJM5tVP957R937+GV4AtC+yXM41PVfgywFUnuNWCq8w'\
             b'C7AXBQl87DVcvLqe19VFnPUhm1VKSbUSCQmglMpqURN5Q9do'\
             b'2jt4bIgnvvAsL0zkpoUZvyxbub3Iy3zBf6Vf6pUv2imL90ur'\
             b'4r+ZV20z5exXDvifJEtlWbzxhRteRvRvA/BS/VSKr4QoHzLH'\
             b'XpHQ/7fHfwMs8lEpPuwWGAAAAABJRU5ErkJggg=='
